"use strict";

var Promise = require("bluebird");

var logger = require('../logger');
var Functions = require('../Functions');

var AbstractCustomMap = require('./AbstractCustomMap');
var Alias = require('./data/Alias');
var CommentDialog = require('../gui/CommentDialog');
var ControlType = require('./ControlType');
var CustomMapOptions = require('./CustomMapOptions');
var GuiConnector = require('../GuiConnector');
var IdentifiedElement = require('./data/IdentifiedElement');
var DataOverlay = require('./data/DataOverlay');
var PointData = require('./data/PointData');
var Reaction = require('./data/Reaction');
var ReferenceGenome = require('./data/ReferenceGenome');
var SecurityError = require('../SecurityError');
var Submap = require('./Submap');

var Bounds = require('./canvas/Bounds');
var Point = require('./canvas/Point');

/**
 * Default constructor.
 *
 * @param options
 *          CustomMapOptions object representing all parameters needed for map
 *          creation
 */
function CustomMap(options) {
  this._controls = [];

  if (!(options instanceof CustomMapOptions)) {
    options = new CustomMapOptions(options);
  }
  AbstractCustomMap.call(this, options.getProject().getModels()[0], options);

  this.registerListenerType("onBioEntityClick");
  this.registerListenerType("onShowOverlay");
  this.registerListenerType("onHideOverlay");
  this.registerListenerType("onBackgroundOverlayChange");

  this._selectedOverlays = [];

  this.customizeGoogleMapView(options.getElement());

  this.createMapChangedCallbacks();

  /**
   *
   * @type {Object.<string, AbstractDbOverlay>}
   */
  this.overlayCollections = {};

  // which submap is active (where user made interaction for the last time)
  this._activeSubmapId = null;

  this.initialized = true;

  // list of reference genomes
  this._referenceGenome = [];

  this.createSubmaps();

  this._dialogs = [];
}

CustomMap.prototype = Object.create(AbstractCustomMap.prototype);
CustomMap.prototype.constructor = CustomMap;

/**
 * This code must be run after the object is created. It requires to download
 * some data via promises.
 *
 * @return {Promise} with empty result
 */
CustomMap.prototype.init = function () {
  var self = this;
  var sessionData = ServerConnector.getSessionData(self.getProject());

  // if we have background overlay stored in the session then restore it
  var mapType = sessionData.getSelectedBackgroundOverlay();
  // if we have user data overlays stored in the session then restore it
  var ids = sessionData.getVisibleOverlays();
  if (mapType !== undefined) {
    ids.push(mapType);
  }

  return Promise.each(ids, function (overlayId) {
    try {
      return self.openDataOverlay(overlayId);
    } catch (e) {
      if (e instanceof SecurityError) {
        logger.debug(e.message);
        sessionData.setSelectedBackgroundOverlay(self.getProject().getDataOverlays()[0].getId());
      } else {
        return Promise.reject(e);
      }
    }
  }).then(function () {
    return ServerConnector.getLogoImg()
  }).then(function (url) {
    var logo2 = self.getControl(ControlType.LOGO_2_IMG);
    if (!/^(f|ht)tps?:\/\//i.test(url)) {
      url = GuiConnector.getImgPrefix() + url;
    }
    logo2.src = url;
  });
};

/**
 * Create submaps.
 */
CustomMap.prototype.createSubmaps = function () {
  this.submaps = [];
  for (var i = 1; i < this.getProject().getModels().length; i++) {
    this.submaps.push(new Submap(this, this.getProject().getModels()[i]));
  }
};

/**
 * Creates logo and put it on the google map.
 */
CustomMap.prototype.createLogo = function () {
  var logoControlDiv2 = Functions.createElement({type: "div", index: 0});
  var logo2 = Functions.createElement({
    type: 'IMG',
    style: "cursor: pointer; width: 80px",
    id: ControlType.LOGO_2_IMG,
    onclick: function () {
      return ServerConnector.getLogoLink().then(function (url) {
        var win = window.open(url, '_blank');
        win.focus();
      }).catch(GuiConnector.alert);
    }
  });
  this.addControl(logo2);

  logoControlDiv2.appendChild(logo2);

  this.getMapCanvas().addLeftBottomControl(logoControlDiv2);

  var logoControlDiv = Functions.createElement({type: "div", style: "padding: 5px", index: 1});
  var logo = Functions.createElement({
    type: "IMG",
    src: GuiConnector.getImgPrefix() + GuiConnector.getLcsbLogoImg(this.bigButtons),
    style: "cursor: pointer",
    id: ControlType.LOGO_IMG,
    onclick: function () {
      var win = window.open('http://wwwen.uni.lu/lcsb/', '_blank');
      win.focus();
    }
  });
  this.addControl(logo);
  logoControlDiv.appendChild(logo);

  this.getMapCanvas().addRightBottomControl(logoControlDiv);
};

/**
 * Clear all AbstractDbOverlay.
 */
CustomMap.prototype.clearDbOverlays = function () {
  var promises = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      promises.push(collection.clear());
    }
  }
  return Promise.all(promises);
};

/**
 * Open data overlay given in the parameter.
 *
 * @param param
 *          identifier or DataOverlay identifying data overlay to open
 */
CustomMap.prototype.openDataOverlay = function (param) {
  logger.debug("Opening data overlay: " + param);

  var self = this;
  var identifier = param;
  if (param instanceof DataOverlay) {
    identifier = param.getId();
  }
  identifier = parseInt(identifier);

  if (isNaN(identifier)) {
    throw new Error("invalid id: " + param);
  }

  var overlayToOpen = null;
  var overlays = self.getProject().getDataOverlays();
  for (var j = 0; j < overlays.length; j++) {
    var overlay = overlays[j];
    if (overlay.getId() === identifier) {
      overlayToOpen = overlay;
    }
  }

  if (overlayToOpen === null) {
    throw new SecurityError("You have no privileges for selected overlay");
  } else {
    return overlayToOpen.init().then(function () {
      if (overlayToOpen.getInputDataAvailable()) {
        if (self._selectedOverlays[identifier] === true) {
          logger.warn("Overlay " + identifier + " already selected");
          return Promise.resolve();
        } else {
          self._selectedOverlays[identifier] = true;
          return self.getVisibleDataOverlays().then(function (visibleDataOverlays) {
            var ids = [];
            for (var i = 0; i < visibleDataOverlays.length; i++) {
              ids.push(visibleDataOverlays[i].getId());
            }
            ServerConnector.getSessionData(self.getProject()).setVisibleOverlays(ids);
            return self.redrawSelectedDataOverlays();
          }).then(function () {
            var backgroundToOpen;
            var overlays = self.getProject().getDataOverlays();
            for (var i = 0; i < overlays.length; i++) {
              var overlay = overlays[i];
              if (overlay.getName().toLowerCase() === "empty") {
                backgroundToOpen = overlay;
              }
            }
            if (backgroundToOpen === undefined) {
              logger.warn("Cannot find empty background overlay");
            } else {
              return self.openDataOverlay(backgroundToOpen.getId());
            }
          });
        }
      } else {
        if (identifier.toString() === self.getMapCanvas().getBackgroundId()) {
          return Promise.resolve();
        } else {
          self.getMapCanvas().setBackgroundId(identifier.toString());

          var submaps = self.getSubmaps();
          for (var i = 0; i < submaps.length; i++) {
            var submap = submaps[i];
            submap.openDataOverlay(identifier);
          }
          return self.callListeners("onBackgroundOverlayChange", identifier);
        }
      }
    }).then(function () {
      return self.callListeners("onShowOverlay", overlayToOpen);
    });
  }
};

/**
 * Returns {DataOverlay} that is used as a background.
 *
 * @returns {DataOverlay}
 */
CustomMap.prototype.getBackgroundDataOverlay = function () {
  var identifier = parseInt(this.getMapCanvas().getBackgroundId());
  var overlays = this.getProject().getDataOverlays();
  for (var i = 0; i < overlays.length; i++) {
    var overlay = overlays[i];
    if (overlay.getId() === identifier) {
      return overlay;
    }
  }
  return null;
};

/**
 * Register AbstractDbOverlay in the map.
 *
 * @param dbOverlay
 *          database overlay to be connected to the map
 */
CustomMap.prototype.registerDbOverlay = function (dbOverlay) {
  var self = this;

  this.overlayCollections[dbOverlay.getName()] = dbOverlay;

  dbOverlay.mapOverlays = {
    ALIAS: [],
    REACTION: [],
    POINT: []
  };

  if (dbOverlay.getName() !== undefined) {
    if (dbOverlay.getName() === "search" || dbOverlay.getName().startsWith("plugin")) {
      dbOverlay.addListener("onSearch", function (e) {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: e.arg.fitBounds
        });
      });
    }
    if (dbOverlay.getName() === "drug" || dbOverlay.getName() === "chemical" || dbOverlay.getName() === "mirna") {
      dbOverlay.addListener("onSearch", function (e) {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: e.arg.fitBounds
        });
      });

      dbOverlay.addListener("onTargetVisibilityChange", function () {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: false
        });
      });
    }
    if (dbOverlay.getName() === "comment") {
      var listener = function () {
        return self.renderOverlayCollection({
          overlayCollection: dbOverlay,
          fitBounds: false
        });
      };
      dbOverlay.addListener("onRefresh", listener);
      dbOverlay.addListener("onClear", listener);
    }
  }

};

/**
 * Refresh comment list.
 *
 * @return promise that is resolved when comment list is refreshed
 */
CustomMap.prototype.refreshComments = function () {
  var self = this;

  var commentDbOverlay = self.getOverlayByName("comment");
  if (commentDbOverlay !== undefined) {
    if (ServerConnector.getSessionData(self.getProject()).getShowComments()) {
      return commentDbOverlay.refresh();
    } else {
      return commentDbOverlay.clear();
    }
  }
  throw new Error("comment DbOverlay not found");
};

/**
 * Turns on/off drawing module on google maps.
 */
CustomMap.prototype.toggleDrawing = function () {
  var submap = this.getSubmapById(this.getActiveSubmapId());
  if (submap !== undefined) {
    submap.getMapCanvas().toggleDrawing();
  } else {
    throw new Error("Cannot find submodel with id: " + this.getActiveSubmapId());
  }
};

CustomMap.prototype.refreshMarkers = function (force) {
  var promises = [];
  logger.debug("Refresh Markers: ");
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      promises.push(this.refreshOverlayMarkers(collection, force));
    }
  }
  return Promise.all(promises);
};

CustomMap.prototype.refreshOverlayMarkers = function (overlay, force) {
  var promises = [];
  var self = this;
  logger.debug("Refresh overlay: " + overlay.name);

  if (!this.isMarkerOptimization() || force) {
    promises.push(self.getMarkerSurfaceCollection().refreshOverlayMarkers(overlay));
    var submaps = self.getSubmaps();
    for (var i = 0; i < submaps.length; i++) {
      if (submaps[i].isInitialized()) {
        promises.push(submaps[i].getMarkerSurfaceCollection().refreshOverlayMarkers(overlay));
      }
    }
  }
  return Promise.all(promises);
};

/**
 * Returns HTML div that is attached to jQuery dialog that should be used for
 * opening submap.
 *
 * @param id
 *          identifier of the submap
 */
CustomMap.prototype.getSubmapDialogDiv = function (id) {
  var dialogDiv = this._dialogs[id];

  if (dialogDiv === undefined) {
    dialogDiv = document.createElement("div");
    dialogDiv.setAttribute("name", "dialog-" + id);

    this._dialogs[id] = dialogDiv;

    $(dialogDiv).dialog({
      autoOpen: false
    });
  }
  return dialogDiv;
};

/**
 * Opens a dialog with a submodel.
 */
CustomMap.prototype.openSubmap = function (id) {
  var self = this;
  var submap = self.getSubmapById(id);
  if (submap === null) {
    throw new Error("Unknown submap for id: " + id);
  } else if (submap !== this) {
    var wasInitialized = submap.isInitialized();

    var dialogDiv = self.getSubmapDialogDiv(id);

    submap.open(dialogDiv);

    return this.openDataOverlay(self.getMapCanvas().getBackgroundId()).then(function () {
      if (!wasInitialized) {
        // now we have to visualize overlays
        var overlayIds = [];

        // get list of overlays
        for (var key in self._selectedOverlays) {
          if (self._selectedOverlays.hasOwnProperty(key) && self._selectedOverlays[key] === true) {
            overlayIds.push(key);
          }
        }

        // show overlays that should be visualized (resize or show them)
        for (var i = 0; i < overlayIds.length; i++) {
          var overlayId = overlayIds[i];
          submap._showSelectedDataOverlay(overlayId, i, overlayIds.length);
        }
        return self.refreshMarkers(true);
      }
    });
  } else {
    return Promise.resolve();
  }

};

CustomMap.prototype.customizeGoogleMapView = function (div) {
  var self = this;

  self.createMapCanvas(div);

  self.createLogo();

  self.getMapCanvas().triggerListeners('resize');
  self.getMapCanvas().triggerListeners('maptypeid_changed');
  self.getMapCanvas().triggerListeners('projection_changed');

  // center map and zoom in to fit into browser window if there is no
  // information about coordinates in the session
  if (ServerConnector.getSessionData(self.getProject()).getCenter(self.getModel()) === undefined &&
    (self.getModel().getDefaultCenterX() === undefined ||
      self.getModel().getDefaultCenterY() === undefined ||
      self.getModel().getDefaultZoomLevel() === undefined ||
      self.getModel().getDefaultCenterX() === null ||
      self.getModel().getDefaultCenterY() === null ||
      self.getModel().getDefaultZoomLevel() === null
    )) {
    var bounds = new Bounds();
    bounds.extend(self.getTopLeft());
    bounds.extend(self.getBottomRight());

    return self.getMapCanvas().fitBounds(bounds);
  }
};

/**
 * Creates listeners for google.maps.Map object that will actualize the data in
 * user session.
 */
CustomMap.prototype.createMapChangedCallbacks = function () {
  this._createMapChangedCallbacks();
  var self = this;
  var sessionData = ServerConnector.getSessionData(self.getProject());

  // listener for changing type of background overlay
  this.getMapCanvas().addListener('maptypeid_changed', function () {
    sessionData.setSelectedBackgroundOverlay(self.getMapCanvas().getBackgroundId());
  });

};

/**
 * Returns submap (or this map) by id.
 *
 * @param identifier
 *          identifier of the submap
 * @returns {AbstractCustomMap} (submap or this map) with given identifier of the model
 */
CustomMap.prototype.getSubmapById = function (identifier) {
  identifier = parseInt(identifier);
  if (this.getId() === identifier) {
    return this;
  }
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].getId() === identifier) {
      return this.submaps[i];
    }
  }
  logger.warn("Cannot find submodel with id: " + identifier);
  return null;
};

CustomMap.prototype.removeSelection = function () {
  var model = this.getSubmapById(this.getActiveSubmapId());
  if (model !== null) {
    model.getMapCanvas().removeSelection();
  } else {
    throw new Error("Cannot find submap with id: " + this.getActiveSubmapId());
  }
};

/**
 * Removes overlay from visualization.
 *
 * @param identifier
 *          identifier of overlay to remove
 *
 */
CustomMap.prototype.hideDataOverlay = function (identifier) {
  var self = this;
  logger.debug("Hiding overlay: " + identifier);

  if (this._selectedOverlays[identifier] !== true) {
    logger.warn("Overlay " + identifier + " is not selected");
    return Promise.resolve();
  } else {
    self._selectedOverlays[identifier] = false;
    return self.redrawSelectedDataOverlays().then(function () {
      return self.getVisibleDataOverlays();
    }).then(function (visibleDataOverlays) {
      var ids = [];
      for (var i = 0; i < visibleDataOverlays.length; i++) {
        ids.push(visibleDataOverlays[i].getId());
      }
      ServerConnector.getSessionData(self.getProject()).setVisibleOverlays(ids);
      return self.redrawSelectedDataOverlays();
    }).then(function () {
      return self.getProject().getDataOverlayById(identifier);
    }).then(function (overlay) {
      return self.callListeners("onHideOverlay", overlay);
    });
  }
};

/**
 * Redrawing selected overlays.
 */
CustomMap.prototype.redrawSelectedDataOverlays = function () {
  logger.debug("Redrawing overlays");
  var self = this;
  return self.getVisibleDataOverlays().then(function (visibleDataOverlays) {
    // show overlays that should be visualized (resize or show them)
    var promises = [];
    for (var i = 0; i < visibleDataOverlays.length; i++) {
      var overlayId = visibleDataOverlays[i].getId();
      if (self.layoutContainsOverlays(overlayId)) {
        // resize element on the map
        promises.push(self.resizeSelectedDataOverlay(overlayId, i, visibleDataOverlays.length));
      } else {
        promises.push(self.showSelectedDataOverlay(overlayId, i, visibleDataOverlays.length));
      }
    }
    return Promise.all(promises).then(function () {
      // hide overlays that were visible
      for (var key in self.selectedLayoutOverlays) {
        if (!self._selectedOverlays.hasOwnProperty(key) || self._selectedOverlays[key] === false) {
          if (self.layoutContainsOverlays(key)) {
            self.hideSelectedLayout(key);
          }
        }
      }
      return self.refreshInfoWindows();
    });
  });
};

/**
 * Hides overlay on the map and all submaps
 *
 * @param overlayId
 *          identifier of a overlay to hide
 */
CustomMap.prototype.hideSelectedLayout = function (overlayId) {
  this._hideSelectedLayout(overlayId);
  for (var i = 0; i < this.submaps.length; i++) {
    this.submaps[i]._hideSelectedLayout(overlayId);
  }
};

/**
 * Resize(refresh) overlay on the map and all submaps. Resizing should be called
 * when number of overlays to visualize change.
 *
 * @param overlayId
 *          identifier of overlay to refresh
 * @param index
 *          position of the overlay in list of overlays that we visualize
 * @param length
 *          number of overlays that we currently visualize
 */
CustomMap.prototype.resizeSelectedDataOverlay = function (overlayId, index, length) {
  logger.debug("Resize overlay: " + overlayId);
  var promises = [];
  promises.push(this._resizeSelectedDataOverlay(overlayId, index, length));
  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._resizeSelectedDataOverlay(overlayId, index, length));
  }
  return Promise.all(promises);
};

/**
 * Show overlay on the map and all submaps.
 *
 * @param overlayId
 *          identifier of overlay to show
 * @param index
 *          position of the overlay in list of overlays that we visualize
 * @param length
 *          number of overlays that we currently visualize
 */
CustomMap.prototype.showSelectedDataOverlay = function (overlayId, index, length) {
  logger.debug("Show overlay: " + overlayId);
  var promises = [];
  promises.push(this._showSelectedDataOverlay(overlayId, index, length));

  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._showSelectedDataOverlay(overlayId, index, length));
  }
  return Promise.all(promises);
};

/**
 * This method checks if the overlay contains any overlays (like AliasSurface or
 * ReactionSurface) that is currently visible on the map.
 *
 * @param overlayId
 *          identifier of the overlay
 * @returns {Boolean}: <code>true</code> if the overlay contains overlays to
 *          visualize, <code>false</code> otherwise
 */
CustomMap.prototype.layoutContainsOverlays = function (overlayId) {

  // first, check top map
  if (this.selectedLayoutOverlays.hasOwnProperty(overlayId) && this.selectedLayoutOverlays[overlayId].length > 0) {
    return true;
  }

  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].isInitialized()) {
      if (this.submaps[i].selectedLayoutOverlays.hasOwnProperty(overlayId)
        && this.submaps[i].selectedLayoutOverlays[overlayId].length > 0) {
        return true;
      }
    }
  }
  return false;
};

/**
 * Refresh content of all {@link AliasInfoWindow} in this map and all submaps.
 */
CustomMap.prototype.refreshInfoWindows = function () {
  var promises = [];
  promises.push(this._refreshInfoWindows());
  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._refreshInfoWindows());
  }
  return Promise.all(promises);
};

/**
 * Opens {@link AliasInfoWindow} for an {@link Alias}.
 *
 * @param alias
 *          {@link Alias}
 * @param marker google.maps.Marker object
 */
CustomMap.prototype.openInfoWindowForAlias = function (alias, marker) {
  logger.debug("Opening info window for alias: " + alias.getId() + ", model: " + alias.getModelId());
  var submap = this.getSubmapById(alias.getModelId());
  return submap._openInfoWindowForAlias(alias, marker);
};

/**
 * Renders markers, lines, etc. for elements highlighted in OverlayCollection.
 *
 * @param params: {overlayCollection, fitBounds}
 *          overlayCollection - to be processed
 *          fitBounds - <code>true</code> if the borders should fit bounds after creating
 *          all elements
 */
CustomMap.prototype.renderOverlayCollection = function (params) {
  var self = this;
  var fitBounds = params.fitBounds;
  var overlayCollection = params.overlayCollection;

  var elements;
  var submaps = self.getSubmaps().concat([self]);

  return overlayCollection.getIdentifiedElements().then(function (identifiedElements) {
    return self.appendElementsPointingToSubmap(identifiedElements, overlayCollection);
  }).then(function (identifiedElements) {
    elements = identifiedElements;
    var promises = [];
    for (var i = 0; i < submaps.length; i++) {
      promises.push(submaps[i].getMarkerSurfaceCollection().renderOverlay(identifiedElements, overlayCollection));
    }
    return Promise.all(promises);
  }).then(function () {
    return Promise.each(elements, function (element) {
      var infoWindow = self.getInfoWindowForIdentifiedElement(element);
      if (infoWindow !== null && infoWindow !== undefined) {
        return self.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray()).then(function () {
          return self.updateInfoWindowForIdentifiedElement(element);
        });
      }
    });
  }).then(function () {
    return self.fetchIdentifiedElements(elements);
  }).then(function (fullElements) {
    var promises = [];
    if (elements.length > 0 && fitBounds) {
      for (var j = 0; j < submaps.length; j++) {
        promises.push(submaps[j].fitBounds(fullElements));
      }
    }
    return Promise.all(promises);
  });
};

CustomMap.prototype.appendElementsPointingToSubmap = function (identifiedElements, overlayDb) {
  var self = this;
  var existingElements = {};
  var i;
  for (i = 0; i < identifiedElements.length; i++) {
    var element = identifiedElements[i];
    existingElements[element.getId() + "_" + element.getIcon() + "_" + element.getType()] = true;
  }
  return Promise.each(identifiedElements, function (ie) {
    return self.getProject().getElementsPointingToSubmap(ie.getModelId()).then(function (elementsPointingToSubmap) {
      for (var j = 0; j < elementsPointingToSubmap.length; j++) {
        var element = elementsPointingToSubmap[j];
        if (!existingElements[element.getId() + "_" + ie.getIcon() + "_" + element.getType()]) {
          existingElements[element.getId() + "_" + ie.getIcon() + "_" + element.getType()] = true;
          var newElement = new IdentifiedElement(elementsPointingToSubmap[j]);
          if (ie.getIcon() === undefined) {
            newElement.setIcon(overlayDb.getIcon(0));
          } else {
            newElement.setIcon(ie.getIcon());
          }

          identifiedElements.push(newElement);
        }
      }
    })
  });
};


/**
 * Opens {@link AbstractInfoWindow} for a marker.
 *
 * @param marker
 *          {AbstractMarker} for which info window should be opened
 * @param element
 *          {IdentifiedElement} for which info window should be opened
 */
CustomMap.prototype.openInfoWindowForIdentifiedElement = function (element, marker) {
  var self = this;
  var submap = self.getSubmapById(element.getModelId());
  logger.debug(element + ": Opening info window");

  // open AliasInfoWindow in a right model
  return submap._openInfoWindowForIdentifiedElement(element, marker).then(function () {
    var infoWindow = submap.returnInfoWindowForIdentifiedElement(element);
    return self.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray());
  });

};

/**
 * Sends requests to download detailed data about element in all
 * {OverlayCollection}.
 *
 * @param element
 *          element for which we want to have detailed information
 * @param {Object.<string,boolean>} general
 */
CustomMap.prototype.retrieveOverlayDetailDataForElement = function (element, general) {
  var promises = [];
  if (general === undefined) {
    logger.warn("general param is undefined!");
    general = {};
  }
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];

      var generalRequest = general[overlayName];
      if (generalRequest === undefined) {
        logger.warn("No information about general overlay request for overlay: "+ overlayName);
        generalRequest = false;
      }
      generalRequest = generalRequest || (!overlay.allowSearchById() && overlay.allowGeneralSearch());

      promises.push(overlay.getDetailDataByIdentifiedElement(element, generalRequest));
    }
  }
  return Promise.all(promises);
};

/**
 * Updates info window identified by element given as a parameter.
 *
 * @param identifiedElement
 *          element for which info window should be updated
 */
CustomMap.prototype.updateInfoWindowForIdentifiedElement = function (identifiedElement) {
  var infoWindow = this.getInfoWindowForIdentifiedElement(identifiedElement);
  if (infoWindow === null || infoWindow === undefined) {
    return Promise.resolve();
  } else {
    return infoWindow.update();
  }
};

/**
 * Returns data from all {OverlayCollection} for a given alias.
 *
 * @param alias
 *          {@link Alias} for which overlay data will be returned
 * @param general
 *          if true then all elements will be returned, if false then only ones
 *          available right now in the overlay
 * @returns {Promise} of the data from all {OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForAlias = function (alias, general) {
  var identifiedElement = new IdentifiedElement(alias);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {OverlayCollection} for a given reaction.
 *
 * @param reaction
 *          {Reaction} for which overlay data will be returned
 * @param general {Array}
 * @returns {Promise} of data from all {OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForReaction = function (reaction, general) {
  var identifiedElement = new IdentifiedElement(reaction);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {OverlayCollection} for a given {PointData}
 *
 * @param general {Array}
 * @param point
 *          {PointData} for which overlay data will be returned
 * @returns {Promise} of data from all {OverlayCollection} for a given
 *          {PointData}
 */
CustomMap.prototype.getOverlayDataForPoint = function (point, general) {
  var identifiedElement = new IdentifiedElement(point);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {OverlayCollection} for element identified by the
 * parameter
 *
 * @param general {Array}
 * @param identifiedElement
 *          {IdentifiedElement} for which overlay data will be returned
 * @returns {Promise} of data from all {OverlayCollection} for a given
 *          {IdentifiedElement}
 */
CustomMap.prototype.getOverlayDataForIdentifiedElement = function (identifiedElement, general) {
  if (general === undefined) {
    logger.warn("general parameter must be defined");
    general = [];
  }
  var promises = [];
  var overlays = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];
      if (overlay.allowGeneralSearch() || overlay.allowSearchById()) {
        var generalFlag = general[overlay.getName()];
        if (generalFlag === undefined) {
          logger.warn("General flag for overlay: " + overlay.getName() + " is not defined, assuming false");
          generalFlag = false;
        }
        overlays.push(overlay);
        promises.push(overlay.getDetailDataByIdentifiedElement(identifiedElement, !overlay.allowSearchById()
          || generalFlag));
      }
    }
  }
  return Promise.all(promises).then(function (values) {
    var result = [];
    for (var i = 0; i < values.length; i++) {
      result.push({
        overlay: overlays[i],
        data: values[i]
      });
    }
    return result;
  });
};

/**
 *
 * @param name
 * @returns {AbstractDbOverlay}
 */
CustomMap.prototype.getOverlayByName = function (name) {
  return this.overlayCollections[name];
};

CustomMap.prototype.getDbOverlays = function () {
  var result = [];
  for (var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      result.push(this.overlayCollections[overlayName]);
    }
  }
  return result;
};

/**
 * Returns {@link AbstractInfoWindow} for element identified by the parameter.
 *
 * @param identifiedElement
 *          {IdentifiedElement} that determines for which element we want
 *          {AbstractInfoWindow}
 * @returns {AbstractInfoWindow} for element identified by the parameter
 */
CustomMap.prototype.getInfoWindowForIdentifiedElement = function (identifiedElement) {
  var model = this.getSubmapById(identifiedElement.modelId);
  var infoWindow = null;
  if (identifiedElement.type === "ALIAS") {
    infoWindow = model.getAliasInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "POINT") {
    infoWindow = model.getPointInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "REACTION") {
    infoWindow = model.getReactionInfoWindowById(identifiedElement.getId());
  } else {
    throw new Error("Unknown type of IdentifiedElement: " + identifiedElement.type);
  }
  return infoWindow;
};

/**
 *
 * @returns {null|number}
 */
CustomMap.prototype.getActiveSubmapId = function () {
  return this._activeSubmapId;
};

CustomMap.prototype.setActiveSubmapId = function (submapId) {
  this._activeSubmapId = submapId;
};

CustomMap.prototype.setActiveSubmapClickCoordinates = function (coordinates) {
  if (!(coordinates instanceof Point)) {
    throw new Error("Coordinates must be provided as Point object, but found: " + coordinates);
  }
  this._activeSubmapCoordinates = coordinates;
};
CustomMap.prototype.getActiveSubmapClickCoordinates = function () {
  return this._activeSubmapCoordinates;
};

CustomMap.prototype.getReferenceGenome = function (type, version) {
  var self = this;
  if (self.getProject().getOrganism() === undefined) {
    return Promise.resolve(new ReferenceGenome(null));
  }

  if (self._referenceGenome[type] === undefined) {
    self._referenceGenome[type] = [];
  }
  if (self._referenceGenome[type][version] === undefined) {
    return ServerConnector.getReferenceGenome({
      type: type,
      version: version,
      organism: self.getProject().getOrganism().getResource()
    }).then(function (genome) {
      self._referenceGenome[type][version] = genome;
      return genome;
    });
  } else {
    return Promise.resolve(self._referenceGenome[type][version]);
  }
};

CustomMap.prototype.getTopMap = function () {
  return this;
};

CustomMap.prototype.getTopOverviewImage = function () {
  return this.getProject().getTopOverviewImage();
};

CustomMap.prototype.addControl = function (element) {
  if (this._controls[element.id] !== undefined) {
    throw new Error("Element with id " + element.id + " already added to controls");
  }
  if (ControlType[element.id] === undefined) {
    throw new Error("Unknown control type: " + element.id);
  }
  this._controls[element.id] = element;
};

CustomMap.prototype.getControl = function (type) {
  if (ControlType[type] === undefined) {
    throw new Error("Unknown control type: " + type);
  }
  return this._controls[type];
};

CustomMap.prototype.setCommentDialog = function (commentDialog) {
  this._commentDialog = commentDialog;
};

CustomMap.prototype.getCommentDialog = function () {
  return this._commentDialog;
};

CustomMap.prototype.setContextMenu = function (contextMenu) {
  this._contextMenu = contextMenu;
};

/**
 *
 * @returns {ContextMenu}
 */
CustomMap.prototype.getContextMenu = function () {
  return this._contextMenu;
};

CustomMap.prototype.setSelectionContextMenu = function (selectionContextMenu) {
  this._selectionContextMenu = selectionContextMenu;
};

CustomMap.prototype.getSelectionContextMenu = function () {
  return this._selectionContextMenu;
};

CustomMap.prototype.openCommentDialog = function () {
  var self = this;
  return ServerConnector.getClosestElementsByCoordinates({
    modelId: this.getActiveSubmapId(),
    coordinates: this.getActiveSubmapClickCoordinates()
  }).then(function (elements) {
    var commentDialog = self.getCommentDialog();
    if (commentDialog === undefined) {
      var div = Functions.createElement({
        type: "div"
      });
      self.getElement().appendChild(div);
      commentDialog = new CommentDialog({
        element: div,
        customMap: self
      });
      self.setCommentDialog(commentDialog);
    }
    self.setCommentDialog(commentDialog);

    return commentDialog.open(elements);
  });
};

CustomMap.prototype.fetchIdentifiedElements = function (elements, complete) {
  var modelIds = [];
  var modelElements = [];
  var i;
  for (i = 0; i < elements.length; i++) {
    var element = elements[i];
    if (modelElements[element.getModelId()] === undefined) {
      modelIds.push(element.getModelId());
      modelElements[element.getModelId()] = [];
    }
    modelElements[element.getModelId()].push(element);
  }

  var promises = [];
  for (i = 0; i < modelIds.length; i++) {
    var modelId = modelIds[i];
    promises.push(this.getSubmapById(modelId).getModel().getByIdentifiedElements(modelElements[modelId], complete));
  }
  var result = [];
  return Promise.all(promises).then(function (data) {
    for (var i = 0; i < data.length; i++) {
      result.push.apply(result, data[i]);
    }
    return result;
  });

};


CustomMap.prototype.setSelectedPolygon = function (polygonData) {
  this._selectedPolygon = polygonData;
};

CustomMap.prototype.getSelectedPolygon = function () {
  return this._selectedPolygon;
};

CustomMap.prototype.getDistance = function (params) {
  var self = this;
  var ie = params.element;
  var model = self.getSubmapById(ie.getModelId()).getModel();
  if (ie.getModelId() !== params.modelId) {
    throw new Error("Element and coordinates are on different maps: " + ie.getModelId() + ", " + params.modelId);
  }
  var x = params.coordinates.x;
  var y = params.coordinates.y;
  var p1 = new Point(x, y);
  return model.getByIdentifiedElement(ie).then(function (element) {
    if (element instanceof Alias) {
      if (element.getX() <= x && element.getX() + element.getWidth() >= x) {
        if (element.getY() <= y && element.getY() + element.getHeight() >= y) {
          return 0;
        } else {
          return Math.min( //
            Math.abs(element.getY() - y), //
            Math.abs(element.getY() + element.getHeight() - y) //
          );
        }
      } else if (element.getY() <= y && element.getY() + element.getHeight() >= y) {
        return Math.min( //
          Math.abs(element.getX() - x), //
          Math.abs(element.getX() + element.getWidth() - x) //
        );
      } else {
        var elementX = element.getX();
        var elementY = element.getY();
        var elementWidth = element.getWidth();
        var elementHeight = element.getHeight();
        return Math.min( //
          Functions.distance(p1, new Point(elementX, y)), //
          Functions.distance(p1, new Point(elementX + elementWidth, elementY)), //
          Functions.distance(p1, new Point(elementX, elementY + elementHeight)), //
          Functions.distance(p1, new Point(elementX + elementWidth, elementY + elementHeight)) //
        );
      }
    } else if (element instanceof Reaction) {
      var distance = Number.POSITIVE_INFINITY;
      var lines = element.getLines();
      for (var i = 0; i < lines.length; i++) {
        distance = Math.min(distance, Functions.distance(p1, lines[i]));
      }
      return distance;
    } else {
      throw new Error("Unknown element type: " + (typeof element));
    }
  });
};

CustomMap.prototype.getSubmaps = function () {
  var submaps = this.submaps;
  if (submaps === undefined) {
    submaps = [];
  }
  return submaps;
};

CustomMap.prototype.destroy = function () {
  var self = this;
  AbstractCustomMap.prototype.destroy.call(self);
  var commentDialog = self.getCommentDialog();
  if (commentDialog !== undefined) {
    commentDialog.destroy();
  }
  var submaps = self.getSubmaps();
  for (var i = 0; i < submaps.length; i++) {
    submaps[i].destroy();
  }
  if (self.getContextMenu() !== undefined) {
    self.getContextMenu().destroy();
  }
  if (self.getSelectionContextMenu() !== undefined) {
    self.getSelectionContextMenu().destroy();
  }
};

/**
 *
 * @returns {Bluebird<DataOverlay[]>}
 */
CustomMap.prototype.getVisibleDataOverlays = function () {
  var dataOverlayPromises = [];

  for (var key in this._selectedOverlays) {
    if (this._selectedOverlays.hasOwnProperty(key) && this._selectedOverlays[key] === true) {
      dataOverlayPromises.push(this.getProject().getDataOverlayById(key));
    }
  }

  return Promise.all(dataOverlayPromises).then(function (overlays) {
    overlays.sort(function (dataOverlay1, dataOverlay2) {
      if (dataOverlay1.getOrder() < dataOverlay2.getOrder())
        return -1;
      if (dataOverlay1.getOrder() > dataOverlay2.getOrder())
        return 1;
      return 0;
    });

    return overlays;
  });
};


module.exports = CustomMap;
